/**
 * Copyright (c) 2023 mol* contributors, licensed under MIT, See LICENSE file for more info.
 *
 * @author Adam Midlik <midlik@gmail.com>
 *
 * Example command-line application generating images of PDB structures
 * Build: npm install --no-save gl jpeg-js pngjs  // these packages are not listed in dependencies for performance reasons
 *        npm run build
 * Run:   node lib/commonjs/examples/image-renderer 1cbs ../outputs_1cbs/
 */

import { ArgumentParser } from 'argparse';
import fs from 'fs';
import path from 'path';
import gl from 'gl';
import pngjs from 'pngjs';
import jpegjs from 'jpeg-js';

import { HeadlessPluginContext } from '../mol-plugin/headless-plugin-context';
import { DefaultPluginSpec } from '../mol-plugin/spec';
import { ExternalModules, defaultCanvas3DParams } from '../mol-plugin/util/headless-screenshot';
import { setFSModule } from '../mol-util/data-source';
import { MembraneOrientationPreset, loadWithUNITMPMembraneRepresentation } from '../tmdet-extension/behavior';
import { TmDetColorThemeProvider } from '../tmdet-extension/tmdet-color-theme';
import { TmDetLabelProvider } from '../tmdet-extension/labeling';


const IMAGE_WIDTH = 800;
const IMAGE_HEIGHT = 800;

setFSModule(fs);

interface Args {
    pdbId: string,
    outDirectory: string
}

function parseArguments(): Args {
    const parser = new ArgumentParser({ description: 'Example command-line application generating images of PDB structures' });
    parser.add_argument('pdbId', { help: 'PDB identifier' });
    parser.add_argument('outDirectory', { help: 'Directory for outputs' });
    const args = parser.parse_args();
    return { ...args };
}

async function main() {
    const args = parseArguments();
    const structureUrl = `https://pdbtm.unitmp.org/api/v1/entry/${args.pdbId}.updated.cif`;
    const regionDescriptorUrl = `https://pdbtm.unitmp.org/api/v1/entry/${args.pdbId}.json`;
    const fileName = `${args.pdbId}.jpg`;
    console.log('PDB ID:', args.pdbId);
    console.log('Structure URL:', structureUrl);
    console.log('Region descriptor URL:', regionDescriptorUrl);
    console.log('Outputs:', args.outDirectory);

    // Create a headless plugin
    const externalModules: ExternalModules = { gl, pngjs, 'jpeg-js': jpegjs };
    const canvasParams = defaultCanvas3DParams();
    canvasParams.camera!.mode = "perspective";
    const plugin = new HeadlessPluginContext(
        externalModules,
        DefaultPluginSpec(),
        { width: IMAGE_WIDTH, height: IMAGE_HEIGHT },
        { canvas: canvasParams }
    );
    await plugin.init();



    //
    // TMDET extension registration section
    //
    let tree = plugin.state.behaviors.build();

    plugin.builders.structure.representation.registerPreset(MembraneOrientationPreset);
    plugin.representation.structure.themes.colorThemeRegistry.add(TmDetColorThemeProvider);
    plugin.managers.lociLabels.addProvider(TmDetLabelProvider);

    await plugin.runTask(plugin.state.behaviors.updateTree(tree, { doNotUpdateCurrent: true, doNotLogTiming: true }));
    await loadWithUNITMPMembraneRepresentation(plugin, {
        structureUrl: structureUrl,
        regionDescriptorUrl: regionDescriptorUrl,
        side1: "Outside" // TODO
    });
    //
    // End of TMDET extension registration section
    //



    // Export images
    fs.mkdirSync(args.outDirectory, { recursive: true });
    setTimeout(async () => {
            plugin.saveImage(path.join(args.outDirectory, fileName));
            // Cleanup
            await plugin.clear();
            plugin.dispose();
        },
        1000
    );

}

main();
