/**
 * Copyright (C) 2022, Protein Bioinformatics Research Group, RCNS
 *
 * Licensed under CC BY-NC 4.0, see LICENSE file for more info.
 *
 * @author Gabor Tusnady <tusnady.gabor@ttk.hu>
 * @author Csongor Gerdan <gerdan.csongor@ttk.hu>
 */

import { Loci } from '../mol-model/loci';
import { StructureElement } from '../mol-model/structure';
import { LociLabel, LociLabelProvider } from '../mol-plugin-state/manager/loci-label';
import { TmDetChainListCache, TmDetDescriptorCache } from './prop';
import { createResidueListsPerChain, getChainAndResidueIds } from './tmdet-color-theme';
import { ChainList, getResidue, ResidueItem } from './types';

const siteLabels = [
    "Side1",
    "Side2",
    "TM alpha",
    "TM beta",
    "TM re-entrant loop",
    "Interfacial Helix",
    "Unknown localization",
    "Membrane Inside"
];
const DefaultResidueLabel = 6; // Unknown localization


export const TmDetLabelProvider: LociLabelProvider = {
    label: (loci: Loci): LociLabel => {
        let labelText = siteLabels[DefaultResidueLabel];

        if (loci.kind == 'element-loci') {
            const unit = loci.elements[0].unit;
            const pdbId = unit.model.entryId;

            const descriptor = TmDetDescriptorCache.get(pdbId);
            if (!descriptor) {
                return labelText;
            }

            let chainList =  TmDetChainListCache.get(pdbId);
            if (!chainList) {
                const tmType = descriptor.additional_entry_annotations.tm_type;
                chainList = createResidueListsPerChain(descriptor.chains, descriptor.side1, tmType);
                TmDetChainListCache.set(pdbId, chainList);
            }

            const location = StructureElement.Loci.getFirstLocation(loci);
            const { chainId, residueId } = getChainAndResidueIds(location!);
            const residue = getResidue(chainList, chainId!, residueId!);
            if (residue) {
                labelText = siteLabels[residue?.siteId];
                let regionText = getRegionText(chainList, chainId!, residue);
                labelText = `${labelText}: ${regionText}`
            }

        }
        return labelText;
    }
}

function getRegionText(chainList: ChainList, chainId: string, residue: ResidueItem): string {
    let value = "Unknown region range";

    const chain = chainList.filter((chain) => chain.chainId === chainId)[0];
    if (chain) {
        // find start of region
        const residues = chain.residues;
        const  authId = parseInt(residue.authId!)
        let previous = residues[residues.length-1];
        for (let i = residues.length-1; i > 0; i--) {
            const current = residues[i];
            const currentId = parseInt(current.authId!);
            // cancel loop when siteId changes
            if (currentId < authId && current.siteId != residue.siteId) {
                break;
            }
            previous = current;
        }
        value = `[ ${previous.authId}`;

        // find end of region
        previous = residues[0];
        for (let current of residues) {
            const currentId = parseInt(current.authId!);
            // cancel loop when siteId changes
            if (authId < currentId && current.siteId !== residue.siteId) {
                break;
            }
            previous = current;
        }
        value = `${value}-${previous.authId} ]`;
    }

    return value;
}
