/**
 * Copyright (C) 2022, Protein Bioinformatics Research Group, RCNS
 *
 * Licensed under CC BY-NC 4.0, see LICENSE file for more info.
 *
 * @author Gabor Tusnady <tusnady.gabor@ttk.hu>
 * @author Csongor Gerdan <gerdan.csongor@ttk.hu>
 */

import { PluginStateObject } from "../mol-plugin-state/objects";
import { StateObject, StateTransformer } from '../mol-state';
import { StateObjectSelector } from "../mol-state/object";

export type PMS = PluginStateObject.Molecule.Structure;

export type PDBTMChain = {
    chain_label: string,
    additional_chain_annotations: {
        type: string,
        num_tm: number
    },
    residues: {
        pdb_res_label?: string,
        aa_type?: string,
        site_data?: { site_id_ref: number, confidence_classification: string }[]
    }[]
};
export type PDBTMRegion = { site: string, auth_ids: number[], color: number[] };
export type PDBTMVec3 = { x: number, y: number, z: number };
type PDBTMTransformationMatrixRow = { x: number, y: number, z: number, t: number };
export type PDBTMTransformationMatrix = {
    rowx: PDBTMTransformationMatrixRow,
    rowy: PDBTMTransformationMatrixRow,
    rowz: PDBTMTransformationMatrixRow
};
export type PDBTMDescriptor = {
    pdb_id: string,
    side1: "Inside"|"Outside"|null,
    chains: PDBTMChain[],
    site: { site_id: number, label: string }[],
    additional_entry_annotations: {
        tm_type: string,
        membrane: {
            normal: PDBTMVec3,
            transformation_matrix: PDBTMTransformationMatrix,
            radius: number
        },
        biomatrix: {
            chain_deletes: string[],
            matrix_list: {
                matrix_id: string,
                apply_to_chain_list: {
                    chain_id: string,
                    new_chain_id: string
                }[],
                transformation_matrix: PDBTMTransformationMatrix
            }[]
        }
    }
};

type StructureComponentType = StateObjectSelector<
        PMS,
        StateTransformer<StateObject<any, StateObject.Type<any>>,
        StateObject<any, StateObject.Type<any>>, any>
    > | undefined;

export type ComponentsType = { polymer: StructureComponentType; ligand: StructureComponentType; water: StructureComponentType };

// for coloring and labeling
export type ResidueItem = {
    authId: string|undefined,
    siteId: number,     // for labels
    siteColorId: number // for site colors
};
export type Chain = {
    chainId: string,
    type: string,
    residues: ResidueItem[]
};
export type ChainList = Chain[];

export function getResidue(chains: ChainList, chainId: string, residueId: string): ResidueItem|undefined {
    let result = undefined;
    const chain = getChain(chains, chainId);
    if (chain) {
        result = chain.residues.filter((res) => res.authId === residueId)[0];
    }

    return result;
}

export function getChain(chains: ChainList, chainId: string): Chain|undefined {
    let result = undefined;
    const chain = chains.filter((chain) => chain.chainId === chainId)[0];
    if (chain) {
        result = chain;
    }
    return result;
}
